require_relative 'logger'
require 'set'

module Awsom
  class SecurityGroup

    include Logger

    DiffStruct = Struct.new(:to_add, :to_remove, keyword_init: true)

    def self.id(name, vpc_id)
      sg = new(name, vpc_id: vpc_id)
      sg.id
    end

    def initialize(name, vpc_id:)
      @name = name
      @vpc_id = vpc_id
    end

    def id
      id = find_id
      raise Error, "specified security_group #{@name} doesn't exist in vpc #{@vpc_id}" if not id
      id
    end

    def description(description)
      @description = description
    end

    def created(&block)
      @desired_rules = Set.new
      instance_eval &block
      @id = find_id || create
      sync
      @id
    end

    private

    def sg
      @sg ||= Aws::EC2::SecurityGroup.new(id)
    end

    def find_id
      filters = [
        { name: "group-name", values: [@name] },
        { name: "vpc-id", values: [@vpc_id] }
      ]
      result = Ec2.describe_security_groups(filters: filters)
      num_results = result.security_groups.size
      if num_results == 1
        result.security_groups.first.group_id
      else
        false
      end
    end

    def create
      resp = Ec2.create_security_group(
        group_name: @name, 
        description: @description,
        vpc_id: @vpc_id
      )
      logger.info "(#{@name}) created security group" 
      resp.group_id
    rescue Aws::Errors::ServiceError, ArgumentError
      error "while creating security_group #{@name}"
    end

    def parse_ports(port)
      case port
      when Integer
        [port, port]
      when Range
        [port.first, port.last]
      when /\A([0-9]+)\z/
        [$1.to_i, $1.to_i]
      when /\A([0-9]+)\-([0-9]+)\z/
        [$1.to_i, $2.to_i]
      else
        raise Error, "invalid port specified #{port}"
      end
    end

    def tcp(port, cidr_ip:)
      (from_port, to_port) = parse_ports(port)
      rule(ip_protocol: "tcp", from_port: from_port, to_port: to_port, cidr_ip: cidr_ip)
    end

    def udp(port, cidr_ip:)
      (from_port, to_port) = parse_ports(port)
      rule(ip_protocol: "udp", from_port: from_port, to_port: to_port, cidr_ip: cidr_ip)
    end

    def icmp(cidr_ip)
      rule(ip_protocol: "icmp", from_port: -1, to_port: -1, cidr_ip: cidr_ip)
    end

    def rule(ip_protocol:, from_port:, to_port:, cidr_ip:)
      rule = {
        ip_protocol: ip_protocol,
        from_port: from_port,
        to_port: to_port,
        cidr_ip: cidr_ip
      }
      @desired_rules << rule
    end

    def current_rules
      current_rules = Set.new 
      sg.ip_permissions.each do |permission|
        rules = permission_to_rules(permission)
        rules.each { |r| current_rules << r }
      end
      current_rules
    end

    def diff(current, desired)
      to_add = desired - current
      to_remove = current - desired
      DiffStruct.new(to_add: to_add, to_remove: to_remove)
    end

    def sync
      current = current_rules
      ingress_diff = diff(current, @desired_rules)
      ingress_diff.to_add.each { |r| authorize_ingress r }
      ingress_diff.to_remove.each { |r| revoke_ingress r }
    end

    def authorize_ingress(rule)
      sg.authorize_ingress rule if not @dry_run
      logger.info "(#{@name}) #{"+".green.bold} allow #{rule[:cidr_ip]} #{rule[:from_port]}-#{rule[:to_port]}"
    rescue Aws::Errors::ServiceError
      error "error authorizing rule #{rule}"
    end

    def revoke_ingress(rule)
      sg.revoke_ingress rule if not @dry_run
      logger.info "(#{@name}) #{"-".red.bold} allow #{rule[:cidr_ip]} #{rule[:from_port]}-#{rule[:to_port]}"
    end

    def permission_to_rules(permission)
      rules = Set.new
      permission.ip_ranges.each do |r|
        rule = {
          ip_protocol: permission.ip_protocol,
          from_port: permission.from_port,
          to_port: permission.to_port,
          cidr_ip: r.cidr_ip
        }
        rules << rule
      end
      rules
    end

  end
end
